import util from 'util';
import F from 'lodash/fp';
import { DockerWrapper, fetchResults, start, stop } from '../micro';

const dumpLog = (log: Array<unknown>) => console.log(util.inspect(log, true, null, true));

describe('Sessions', () => {
  let log: Array<unknown> = [];
  let docker: DockerWrapper;

  const logContains = (ev: unknown) => F.some(F.isMatch(ev as object), log);

  beforeAll(() => {
    browser.call(() => {
      return start().then((container) => {
        docker = container;
      });
    });
    browser.url('/index.html');
    browser.setCookies({ name: 'container', value: docker.url });
    browser.url('/session-integration.html');
    browser.pause(6000); // Time for requests to get written
    browser.call(() =>
      fetchResults(docker.url).then((result) => {
        log = result;
      })
    );
  });

  afterAll(() => {
    browser.call(() => {
      return stop(docker.container);
    });
  });

  it('should count sessions using cookies', () => {
    expect(
      logContains({
        event: {
          event: 'page_view',
          name_tracker: 'cookieSessionTracker',
          domain_sessionidx: 2,
        },
      })
    ).toBe(true);
  });

  it('should count sessions using local storage', () => {
    expect(
      logContains({
        event: {
          event: 'page_view',
          name_tracker: 'localStorageSessionTracker',
          domain_sessionidx: 2,
        },
      })
    ).toBe(true);
  });

  it('should count sessions using anonymousSessionTracking', () => {
    expect(
      logContains({
        event: {
          event: 'page_view',
          name_tracker: 'anonymousSessionTracker',
          domain_sessionidx: 2,
        },
      })
    ).toBe(true);
  });

  it('should only increment domain_sessionidx outside of session timeout (local storage)', () => {
    const withSingleVid = (ev: unknown) =>
      F.get('event.name_tracker', ev) === 'localStorageSessionTracker' && F.get('event.domain_sessionidx', ev) === 1;

    expect(F.size(F.filter(withSingleVid, log))).toBe(2);
    expect(F.size(F.filter((e) => F.get('event.name_tracker', e) === 'localStorageSessionTracker', log))).toBe(3);
  });

  it('should only increment domain_sessionidx outside of session timeout (anonymous session tracking)', () => {
    const withSingleVid = (ev: unknown) =>
      F.get('event.name_tracker', ev) === 'anonymousSessionTracker' && F.get('event.domain_sessionidx', ev) === 1;

    expect(F.size(F.filter(withSingleVid, log))).toBe(2);
    expect(F.size(F.filter((e) => F.get('event.name_tracker', e) === 'anonymousSessionTracker', log))).toBe(3);
  });

  it('should only increment domain_sessionidx outside of session timeout (cookie storage)', () => {
    const withSingleVid = (ev: unknown) =>
      F.get('event.name_tracker', ev) === 'cookieSessionTracker' && F.get('event.domain_sessionidx', ev) === 1;

    expect(F.size(F.filter(withSingleVid, log))).toBe(2);
    expect(F.size(F.filter((e) => F.get('event.name_tracker', e) === 'cookieSessionTracker', log))).toBe(3);
  });
});
